import torch

import penn


###############################################################################
# Batch sampler
###############################################################################


def sampler(dataset, partition):
    """Create batch index sampler"""
    # Get sampler indices
    indices = (
        dataset.voiced_indices() if penn.VOICED_ONLY and partition == 'train'
        else list(range(len(dataset))))

    # Maybe use distributed sampler for training
    if partition == 'train':
        return Sampler(indices)

    # Possibly deterministic random sampler for validation
    elif partition == 'valid':
        return Sampler(indices)

    # Sample test data sequentially
    elif partition == 'test':
        return torch.utils.data.SequentialSampler(dataset)

    else:
        raise ValueError(f'Partition {partition} is not implemented')


###############################################################################
# Custom samplers
###############################################################################


class Sampler:

    def __init__(self, indices):
        self.indices = indices
        self.epoch = 0

    def __iter__(self):
        generator = torch.Generator()
        generator.manual_seed(penn.RANDOM_SEED + self.epoch)
        for i in torch.randperm(len(self.indices), generator=generator):
            yield self.indices[i]

    def __len__(self):
        return len(self.indices)

    def set_epoch(self, epoch):
        self.epoch = epoch



import contextlib
import functools
import math
import time
from pathlib import Path
from typing import List, Optional, Tuple

import huggingface_hub
import torch
import torch.multiprocessing as mp
import torchaudio
import torchutil

import penn


###############################################################################
# Pitch and periodicity estimation
###############################################################################


def from_audio(
        audio: torch.Tensor,
        sample_rate: int = penn.SAMPLE_RATE,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Optional[Path] = None,
        batch_size: Optional[int] = None,
        center: str = 'half-window',
        interp_unvoiced_at: Optional[float] = None,
        gpu: Optional[int] = None
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Perform pitch and periodicity estimation

    Args:
        audio: The audio to extract pitch and periodicity from
        sample_rate: The audio sample rate
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        center: Padding options. One of ['half-window', 'half-hop', 'zero'].
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on

    Returns:
        pitch: torch.tensor(
            shape=(1, int(samples // penn.seconds_to_sample(hopsize))))
        periodicity: torch.tensor(
            shape=(1, int(samples // penn.seconds_to_sample(hopsize))))
    """
    pitch, periodicity = [], []

    # Preprocess audio
    for frames in preprocess(
        audio,
        sample_rate,
        hopsize,
        batch_size,
        center
    ):

        # Copy to device
        with torchutil.time.context('copy-to'):
            frames = frames.to('cpu' if gpu is None else f'cuda:{gpu}')

        # Infer
        logits = infer(frames, checkpoint).detach()

        # Postprocess
        with torchutil.time.context('postprocess'):
            result = postprocess(logits, fmin, fmax)
            pitch.append(result[1])
            periodicity.append(result[2])

    # Concatenate results
    pitch, periodicity = torch.cat(pitch, 1), torch.cat(periodicity, 1)

    # Maybe interpolate unvoiced regions
    if interp_unvoiced_at is not None:
        pitch = penn.voicing.interpolate(
            pitch,
            periodicity,
            interp_unvoiced_at)

    return pitch, periodicity


def from_file(
        file: Path,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Optional[Path] = None,
        batch_size: Optional[int] = None,
        center: str = 'half-window',
        interp_unvoiced_at: Optional[float] = None,
        gpu: Optional[int] = None
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Perform pitch and periodicity estimation from audio on disk

    Args:
        file: The audio file
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        center: Padding options. One of ['half-window', 'half-hop', 'zero'].
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on

    Returns:
        pitch: torch.tensor(shape=(1, int(samples // hopsize)))
        periodicity: torch.tensor(shape=(1, int(samples // hopsize)))
    """
    # Load audio
    with torchutil.time.context('load'):
        audio, sample_rate = torchaudio.load(file)

    # Inference
    return from_audio(
        audio,
        sample_rate,
        hopsize,
        fmin,
        fmax,
        checkpoint,
        batch_size,
        center,
        interp_unvoiced_at,
        gpu)


def from_file_to_file(
        file: Path,
        output_prefix: Optional[Path] = None,
        hopsize: float = penn.HOPSIZE_SECONDS,
        fmin: float = penn.FMIN,
        fmax: float = penn.FMAX,
        checkpoint: Optional[Path] = None,
        batch_size: Optional[int] = None,
        center: str = 'half-window',
        interp_unvoiced_at: Optional[float] = None,
        gpu: Optional[int] = None
) -> None:
    """Perform pitch and periodicity estimation from audio on disk and save

    Args:
        file: The audio file
        output_prefix: The file to save pitch and periodicity without extension
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        center: Padding options. One of ['half-window', 'half-hop', 'zero'].
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        gpu: The index of the gpu to run inference on
    """
    # Inference
    pitch, periodicity = from_file(
        file,
        hopsize,
        fmin,
        fmax,
        checkpoint,
        batch_size,
        center,
        interp_unvoiced_at,
        gpu)

    # Move to cpu
    with torchutil.time.context('copy-from'):
        pitch, periodicity = pitch.cpu(), periodicity.cpu()

    # Save to disk
    with torchutil.time.context('save'):

        # Maybe use same filename with new extension
        if output_prefix is None:
            output_prefix = file.parent / file.stem

        # Save
        torch.save(pitch, f'{output_prefix}-pitch.pt')
        torch.save(periodicity, f'{output_prefix}-periodicity.pt')


def from_files_to_files(
    files: List[Path],
    output_prefixes: Optional[List[Path]] = None,
    hopsize: float = penn.HOPSIZE_SECONDS,
    fmin: float = penn.FMIN,
    fmax: float = penn.FMAX,
    checkpoint: Optional[Path] = None,
    batch_size: Optional[int] = None,
    center: str = 'half-window',
    interp_unvoiced_at: Optional[float] = None,
    num_workers: int = penn.NUM_WORKERS,
    gpu: Optional[int] = None
) -> None:
    """Perform pitch and periodicity estimation from files on disk and save

    Args:
        files: The audio files
        output_prefixes: Files to save pitch and periodicity without extension
        hopsize: The hopsize in seconds
        fmin: The minimum allowable frequency in Hz
        fmax: The maximum allowable frequency in Hz
        checkpoint: The checkpoint file
        batch_size: The number of frames per batch
        center: Padding options. One of ['half-window', 'half-hop', 'zero'].
        interp_unvoiced_at: Specifies voicing threshold for interpolation
        num_workers: Number of CPU threads for async data I/O
        gpu: The index of the gpu to run inference on
    """
    # Maybe use default output filenames
    if output_prefixes is None:
        output_prefixes = [file.parent / file.stem for file in files]

    # Single-threaded
    if num_workers == 0:

        # Iterate over files
        for i, (file, output_prefix) in torchutil.iterator(
            enumerate(zip(files, output_prefixes)),
            f'{penn.CONFIG}',
            total=len(files)
        ):

            # Infer
            from_file_to_file(
                file,
                output_prefix,
                hopsize,
                fmin,
                fmax,
                checkpoint,
                batch_size,
                center,
                interp_unvoiced_at,
                gpu)

    # Multi-threaded
    else:

        # Initialize multi-threaded dataloader
        loader = inference_loader(
            files,
            hopsize,
            batch_size,
            center,
            int(math.ceil(num_workers / 2)))

        # Maintain file correspondence
        output_prefixes = {
            file: output_prefix
            for file, output_prefix in zip(files, output_prefixes)}

        # Setup multiprocessing
        futures = []
        pool = mp.get_context('spawn').Pool(max(1, num_workers // 2))

        # Setup progress bar
        progress = torchutil.iterator(
            range(len(files)),
            penn.CONFIG,
            total=len(files))

        try:

            # Track residual to fill batch
            residual_files = []
            residual_frames = torch.zeros((0, 1, 1024))
            residual_lengths = torch.zeros((0,), dtype=torch.long)

            # Iterate over data
            pitch, periodicity = torch.zeros((1, 0)), torch.zeros((1, 0))

            for frames, lengths, input_files in loader:

                # Prepend residual
                if residual_files:
                    frames = torch.cat((residual_frames, frames), dim=0)
                    lengths = torch.cat((residual_lengths, lengths))
                    input_files = residual_files + input_files

                i = 0
                while batch_size is None or i + batch_size <= len(frames):

                    # Copy to device
                    size = len(frames) if batch_size is None else batch_size
                    batch_frames = frames[i:i + size].to(
                        'cpu' if gpu is None else f'cuda:{gpu}')

                    # Infer
                    logits = infer(batch_frames, checkpoint).detach()

                    # Postprocess
                    results = postprocess(logits, fmin, fmax)

                    # Append to residual
                    pitch = torch.cat((pitch, results[1].cpu()), dim=1)
                    periodicity = torch.cat(
                        (periodicity, results[2].cpu()),
                        dim=1)

                    i += len(batch_frames)

                # Save to disk
                j, k = 0, 0
                for length, file in zip(lengths, input_files):

                    # Slice and save in another process
                    if j + length <= pitch.shape[-1]:
                        futures.append(
                            pool.apply_async(
                                save_worker,
                                args=(
                                    output_prefixes[file],
                                    pitch[:, j:j + length],
                                    periodicity[:, j:j + length],
                                    interp_unvoiced_at)))
                        while len(futures) > 100:
                            futures = [f for f in futures if not f.ready()]
                            time.sleep(.1)
                        j += length
                        k += 1
                        progress.update()
                    else:
                        break

                # Setup residual for next iteration
                pitch = pitch[:, j:]
                periodicity = periodicity[:, j:]
                residual_files = input_files[k:]
                residual_lengths = lengths[k:]
                residual_frames = frames[i:]

            # Handle final files
            if residual_frames.numel():

                # Copy to device
                batch_frames = residual_frames.to(
                    'cpu' if gpu is None else f'cuda:{gpu}')

                # Infer
                logits = infer(batch_frames, checkpoint).detach()

                # Postprocess
                results = postprocess(logits, fmin, fmax)

                # Append to residual
                pitch = torch.cat((pitch, results[1].cpu()), dim=1)
                periodicity = torch.cat((periodicity, results[2].cpu()), dim=1)

                # Save
                i = 0
                for length, file in zip(residual_lengths, residual_files):

                    # Slice and save in another process
                    if i + length <= pitch.shape[-1]:
                        futures.append(
                            pool.apply_async(
                                save_worker,
                                args=(
                                    output_prefixes[file],
                                    pitch[:, i:i + length],
                                    periodicity[:, i:i + length],
                                    interp_unvoiced_at)))
                        while len(futures) > 100:
                            futures = [f for f in futures if not f.ready()]
                            time.sleep(.1)
                        i += length
                        progress.update()

            # Wait
            for future in futures:
                future.wait()

        finally:

            # Shutdown multiprocessing
            pool.close()
            pool.join()

            # Close progress bar
            progress.close()


###############################################################################
# Inference pipeline stages
###############################################################################


def infer(frames, checkpoint=None):
    """Forward pass through the model"""
    # Time model loading
    with torchutil.time.context('model'):

        # Load and cache model
        if (
            not hasattr(infer, 'model') or
            infer.checkpoint != checkpoint or
            infer.device != frames.device
        ):

            # Initialize model
            model = penn.Model()

            # Maybe download from HuggingFace
            if checkpoint is None:
                checkpoint = huggingface_hub.hf_hub_download(
                    'maxrmorrison/fcnf0-plus-plus',
                    'fcnf0++.pt')
                infer.checkpoint = None
            else:
                infer.checkpoint = checkpoint
            checkpoint = torch.load(checkpoint, map_location='cpu')

            # Load from disk
            model.load_state_dict(checkpoint['model'])
            infer.device = frames.device

            # Move model to correct device (no-op if devices are the same)
            infer.model = model.to(frames.device)

    # Time inference
    with torchutil.time.context('infer'):

        # Prepare model for inference
        with inference_context(infer.model):

            # Infer
            logits = infer.model(frames)

        # If we're benchmarking, make sure inference finishes within timer
        if penn.BENCHMARK and logits.device.type == 'cuda':
            torch.cuda.synchronize(logits.device)

        return logits


def postprocess(logits, fmin=penn.FMIN, fmax=penn.FMAX):
    """Convert model output to pitch and periodicity"""
    # Turn off gradients
    with torch.inference_mode():

        # Convert frequency range to pitch bin range
        minidx = penn.convert.frequency_to_bins(torch.tensor(fmin))
        maxidx = penn.convert.frequency_to_bins(
            torch.tensor(fmax),
            torch.ceil)

        # Remove frequencies outside of allowable range
        logits[:, :minidx] = -float('inf')
        logits[:, maxidx:] = -float('inf')

        # Decode pitch from logits
        if penn.DECODER == 'argmax':
            bins, pitch = penn.decode.argmax(logits)
        elif penn.DECODER.startswith('viterbi'):
            bins, pitch = penn.decode.viterbi(logits)
        elif penn.DECODER == 'local_expected_value':
            bins, pitch = penn.decode.local_expected_value(logits)
        else:
            raise ValueError(f'Decoder method {penn.DECODER} is not defined')

        # Decode periodicity from logits
        if penn.PERIODICITY == 'entropy':
            periodicity = penn.periodicity.entropy(logits)
        elif penn.PERIODICITY == 'max':
            periodicity = penn.periodicity.max(logits)
        elif penn.PERIODICITY == 'sum':
            periodicity = penn.periodicity.sum(logits)
        else:
            raise ValueError(
                f'Periodicity method {penn.PERIODICITY} is not defined')

        return bins.T, pitch.T, periodicity.T


def preprocess(
    audio,
    sample_rate=penn.SAMPLE_RATE,
    hopsize=penn.HOPSIZE_SECONDS,
    batch_size=None,
    center='half-window'):
    """Convert audio to model input"""
    # Get number of frames
    total_frames = expected_frames(
        audio.shape[-1],
        sample_rate,
        hopsize,
        center)

    # Maybe resample
    if sample_rate != penn.SAMPLE_RATE:
        audio = resample(audio, sample_rate)

    # Maybe pad audio
    hopsize = penn.convert.seconds_to_samples(hopsize)
    if center in ['half-hop', 'zero']:
        if center == 'half-hop':
            padding = int((penn.WINDOW_SIZE - hopsize) / 2)
        else:
            padding = int(penn.WINDOW_SIZE / 2)
        audio = torch.nn.functional.pad(
            audio,
            (padding, padding),
            mode='reflect')

    # Integer hopsizes permit a speedup using torch.unfold
    if isinstance(hopsize, int) or hopsize.is_integer():
        hopsize = int(round(hopsize))
        start_idxs = None

    else:

        # Find start indices
        start_idxs = torch.round(
            torch.tensor([hopsize * i for i in range(total_frames + 1)])
        ).int()

    # Default to running all frames in a single batch
    batch_size = total_frames if batch_size is None else batch_size

    # Generate batches
    for i in range(0, total_frames, batch_size):

        # Size of this batch
        batch = min(total_frames - i, batch_size)

        # Fast implementation for integer hopsizes
        if start_idxs is None:

            # Batch indices
            start = i * hopsize
            end = start + int((batch - 1) * hopsize) + penn.WINDOW_SIZE
            end = min(end, audio.shape[-1])
            batch_audio = audio[:, start:end]

            # Maybe pad to a single frame
            if end - start < penn.WINDOW_SIZE:
                padding = penn.WINDOW_SIZE - (end - start)

                # Handle multiple of hopsize
                remainder = (end - start) % hopsize
                if remainder:
                    padding += end - start - hopsize

                # Pad
                batch_audio = torch.nn.functional.pad(
                    batch_audio,
                    (0, padding))

            # Slice and chunk audio
            frames = torch.nn.functional.unfold(
                batch_audio[:, None, None],
                kernel_size=(1, penn.WINDOW_SIZE),
                stride=(1, hopsize)).permute(2, 0, 1)

        # Slow implementation for floating-point hopsizes
        else:

            # Allocate frames
            frames = torch.zeros(batch, 1, penn.WINDOW_SIZE)

            # Fill each frame with a window starting at the start index
            for j in range(batch):
                start = start_idxs[i + j]
                end = min(start + penn.WINDOW_SIZE, audio.shape[-1])
                frames[j, :, : end - start] = audio[:, start:end]

        yield frames


###############################################################################
# Inference acceleration
###############################################################################


def inference_collate(batch):
    frames, lengths, files = zip(*batch)
    return (
        torch.cat(frames, dim=0),
        torch.tensor(lengths, dtype=torch.long),
        files)


def inference_loader(
    files,
    hopsize=penn.HOPSIZE_SECONDS,
    batch_size=None,
    center='half-window',
    num_workers=penn.NUM_WORKERS // 2
):
    dataset = InferenceDataset(files, hopsize, batch_size, center)
    return torch.utils.data.DataLoader(
        dataset,
        batch_sampler=InferenceSampler(dataset),
        num_workers=num_workers,
        collate_fn=inference_collate)


def save_worker(prefix, pitch, periodicity, interp_unvoiced_at=None):
    """Save pitch and periodicity to disk"""
    # Maybe interpolate unvoiced regions
    if interp_unvoiced_at is not None:
        pitch = penn.voicing.interpolate(
            pitch,
            periodicity,
            interp_unvoiced_at)

    # Save
    Path(prefix).parent.mkdir(exist_ok=True, parents=True)
    torch.save(pitch, f'{prefix}-pitch.pt')
    torch.save(periodicity, f'{prefix}-periodicity.pt')

    # Clean-up
    del pitch
    del periodicity


class InferenceDataset(torch.utils.data.Dataset):

    def __init__(
        self,
        files,
        hopsize=penn.HOPSIZE_SECONDS,
        batch_size=None,
        center='half-window'):
        self.files = files
        self.batch_size = batch_size
        self.lengths = []
        for file in files:
            info = torchaudio.info(file)
            self.lengths.append(
                expected_frames(
                    info.num_frames,
                    info.sample_rate,
                    hopsize,
                    center))
        self.preprocess_fn = functools.partial(
            preprocess,
            hopsize=hopsize,
            batch_size=batch_size,
            center=center)

    def __getitem__(self, index):
        frames = torch.cat(
            [
                frame for frame in
                self.preprocess_fn(*torchaudio.load(self.files[index]))
            ],
            dim=0)
        return frames, self.lengths[index], self.files[index]

    def __len__(self):
        return len(self.files)


class InferenceSampler(torch.utils.data.Sampler):

    def __init__(self, dataset):
        self.dataset = dataset

    def __iter__(self):
        return iter(self.batch)

    def __len__(self):
        return len(self.batch)

    @functools.cached_property
    def batch(self):
        count = 0
        batch, batches = [], []
        for i, length in enumerate(self.dataset.lengths):
            batch.append(i)
            if self.dataset.batch_size is None:
                batches.append(batch)
                batch = []
            else:
                count += length
                if count >= self.dataset.batch_size:
                    batches.append(batch)
                    batch = []
                    count = 0
        if batch:
            batches.append(batch)
        return batches


###############################################################################
# Utilities
###############################################################################


def cents(a, b):
    """Compute pitch difference in cents"""
    return penn.OCTAVE * torch.log2(a / b)


def expected_frames(samples, sample_rate, hopsize, center):
    """Compute expected number of output frames"""
    # Calculate expected number of frames
    hopsize_resampled = penn.convert.seconds_to_samples(
        hopsize,
        sample_rate)
    if center == 'half-window':
        window_size_resampled = \
            penn.WINDOW_SIZE / penn.SAMPLE_RATE * sample_rate
        samples = samples - (window_size_resampled - hopsize_resampled)
    elif center == 'half-hop':
        samples = samples
    elif center == 'zero':
        samples = samples + hopsize_resampled
    else:
        raise ValueError(f'Unknown center sample {center}')
    return max(1, int(samples / hopsize_resampled))


@contextlib.contextmanager
def inference_context(model):
    device_type = next(model.parameters()).device.type

    # Prepare model for evaluation
    model.eval()

    # Turn off gradient computation
    with torch.inference_mode():

        if device_type == 'cuda':
            with torch.autocast(device_type):
                yield
        else:
            yield

    # Prepare model for training
    model.train()


def interpolate(x, xp, fp):
    """1D linear interpolation for monotonically increasing sample points"""
    # Handle edge cases
    if xp.shape[-1] == 0:
        return x
    if xp.shape[-1] == 1:
        return torch.full(
            x.shape,
            fp.squeeze(),
            device=fp.device,
            dtype=fp.dtype)

    # Get slope and intercept using right-side first-differences
    m = (fp[:, 1:] - fp[:, :-1]) / (xp[:, 1:] - xp[:, :-1])
    b = fp[:, :-1] - (m.mul(xp[:, :-1]))

    # Get indices to sample slope and intercept
    indicies = torch.sum(torch.ge(x[:, :, None], xp[:, None, :]), -1) - 1
    indicies = torch.clamp(indicies, 0, m.shape[-1] - 1)
    line_idx = torch.linspace(
        0,
        indicies.shape[0],
        1,
        device=indicies.device).to(torch.long).expand(indicies.shape)

    # Interpolate
    return m[line_idx, indicies].mul(x) + b[line_idx, indicies]


def normalize(frames):
    """Normalize audio frames to have mean zero and std dev one"""
    # Mean-center
    frames -= frames.mean(dim=2, keepdim=True)

    # Scale
    frames /= torch.max(
        torch.tensor(1e-10, device=frames.device),
        frames.std(dim=2, keepdim=True))

    return frames


def resample(audio, sample_rate, target_rate=penn.SAMPLE_RATE):
    """Perform audio resampling"""
    if sample_rate == target_rate:
        return audio
    resampler = torchaudio.transforms.Resample(sample_rate, target_rate)
    resampler = resampler.to(audio.device)
    return resampler(audio)



import penn


###############################################################################
# Test core.py
###############################################################################


def test_infer(audio):
    """Test that inference produces the correct shape"""
    pitch, periodicity = penn.from_audio(
        audio,
        penn.SAMPLE_RATE,
        center='half-hop')
    shape = (1, audio.shape[1] // penn.HOPSIZE)
    assert pitch.shape == periodicity.shape == shape


def test_infer_stereo(audio_stereo):
    """Test that inference on stereo audio produces the correct shape"""
    pitch, periodicity = penn.from_audio(
        audio_stereo,
        penn.SAMPLE_RATE,
        center='half-hop')
    shape = (1, audio_stereo.shape[1] // penn.HOPSIZE)
    assert pitch.shape == periodicity.shape == shape



import penn

import librosa
import torch


###############################################################################
# Test convert.py
###############################################################################


def test_convert_frequency_to_midi():
    """Test that conversion from Hz to MIDI matches librosa implementation"""
    sample_data = torch.tensor([110.0, 220.0, 440.0, 500.0, 880.0])

    # Convert
    penn_midi = penn.convert.frequency_to_midi(sample_data)
    librosa_midi = librosa.hz_to_midi(sample_data.numpy())

    # Compare
    assert torch.allclose(
        penn_midi,
        torch.tensor(librosa_midi, dtype=torch.float32))


def test_convert_midi_to_frequency():
    """Test that conversion from MIDI to Hz matches librosa implementation"""
    sample_data = torch.tensor([45.0, 57.0, 69.0, 71.2131, 81.0])

    # Convert
    penn_frequency = penn.convert.midi_to_frequency(sample_data)
    librosa_frequency = librosa.midi_to_hz(sample_data.numpy())

    # Compare
    assert torch.allclose(
        penn_frequency,
        torch.tensor(librosa_frequency, dtype=torch.float32))



import torch
import torchutil

import penn


###############################################################################
# Constants
###############################################################################


# Evaluation threshold for RPA and RCA
THRESHOLD = 50  # cents


###############################################################################
# Aggregate metric
###############################################################################


class Metrics:

    def __init__(self):
        self.accuracy = torchutil.metrics.Accuracy()
        self.f1 = F1()
        self.loss = Loss()
        self.pitch_metrics = PitchMetrics()

    def __call__(self):
        return (
            {
                'accuracy': self.accuracy(),
                'loss': self.loss()
            } |
            self.f1() |
            self.pitch_metrics())

    def update(self, logits, bins, target, voiced):
        # Detach from graph
        logits = logits.detach()

        # Update loss
        self.loss.update(logits[:, :penn.PITCH_BINS], bins.T)

        # Decode bins, pitch, and periodicity
        with torchutil.time.context('decode'):
            predicted, pitch, periodicity = penn.postprocess(logits)

        # Update bin accuracy
        self.accuracy.update(predicted[voiced], bins[voiced])

        # Update pitch metrics
        self.pitch_metrics.update(pitch, target, voiced)

        # Update periodicity metrics
        self.f1.update(periodicity, voiced)

    def reset(self):
        self.accuracy.reset()
        self.f1.reset()
        self.loss.reset()
        self.pitch_metrics.reset()


class PitchMetrics:

    def __init__(self):
        self.l1 = L1()
        self.rca = RCA()
        self.rmse = RMSE()
        self.rpa = RPA()

    def __call__(self):
        return {
            'l1': self.l1(),
            'rca': self.rca(),
            'rmse': self.rmse(),
            'rpa': self.rpa()}

    def update(self, pitch, target, voiced):
        # Mask unvoiced
        pitch, target = pitch[voiced], target[voiced]

        # Update metrics
        self.l1.update(pitch, target)
        self.rca.update(pitch, target)
        self.rmse.update(pitch, target)
        self.rpa.update(pitch, target)

    def reset(self):
        self.l1.reset()
        self.rca.reset()
        self.rmse.reset()
        self.rpa.reset()


###############################################################################
# Individual metrics
###############################################################################


class F1:

    def __init__(self, thresholds=None):
        if thresholds is None:
            thresholds = sorted(list(set(
                [2 ** -i for i in range(1, 11)] +
                [.1 * i for i in range(10)])))
        self.thresholds = thresholds
        self.precision = [
            torchutil.metrics.Precision() for _ in range(len(thresholds))]
        self.recall = [
            torchutil.metrics.Recall() for _ in range(len(thresholds))]

    def __call__(self):
        result = {}
        for threshold, precision, recall in zip(
            self.thresholds,
            self.precision,
            self.recall
        ):
            precision = precision()
            recall = recall()
            try:
                f1 = 2 * precision * recall / (precision + recall)
            except ZeroDivisionError:
                f1 = 0.
            result |= {
                f'f1-{threshold:.6f}': f1,
                f'precision-{threshold:.6f}': precision,
                f'recall-{threshold:.6f}': recall}
        return result

    def update(self, periodicity, voiced):
        for threshold, precision, recall in zip(
            self.thresholds,
            self.precision,
            self.recall
        ):
            predicted = penn.voicing.threshold(periodicity, threshold)
            precision.update(predicted, voiced)
            recall.update(predicted, voiced)

    def reset(self):
        """Reset the F1 score"""
        for precision, recall in zip(self.precision, self.recall):
            precision.reset()
            recall.reset()


class L1(torchutil.metrics.L1):
    """L1 pitch distance in cents"""
    def update(self, predicted, target):
        super().update(
            penn.OCTAVE * torch.log2(predicted),
            penn.OCTAVE * torch.log2(target))


class Loss(torchutil.metrics.Average):
    """Batch-updating loss"""
    def update(self, logits, bins):
        super().update(penn.loss(logits, bins), bins.shape[0])


class RCA(torchutil.metrics.Average):
    """Raw chroma accuracy"""
    def update(self, predicted, target):
        # Compute pitch difference in cents
        difference = penn.cents(predicted, target)

        # Forgive octave errors
        difference[difference > (penn.OCTAVE - THRESHOLD)] -= penn.OCTAVE
        difference[difference < -(penn.OCTAVE - THRESHOLD)] += penn.OCTAVE

        # Count predictions that are within 50 cents of target
        super().update(
            (torch.abs(difference) < THRESHOLD).sum(),
            predicted.numel())


class RMSE(torchutil.metrics.RMSE):
    """Root mean square error of pitch distance in cents"""
    def update(self, predicted, target):
        super().update(
            penn.OCTAVE * torch.log2(predicted),
            penn.OCTAVE * torch.log2(target))


class RPA(torchutil.metrics.Average):
    """Raw prediction accuracy"""
    def update(self, predicted, target):
        difference = penn.cents(predicted, target)
        super().update(
            (torch.abs(difference) < THRESHOLD).sum(),
            predicted.numel())



import argparse
import shutil
from pathlib import Path

import torchutil

import penn


###############################################################################
# Entry point
###############################################################################


def main(config, datasets, gpu):
    # Create output directory
    directory = penn.RUNS_DIR / config.stem
    directory.mkdir(parents=True, exist_ok=True)

    # Save configuration
    shutil.copyfile(config, directory / config.name)

    # Train
    penn.train(datasets, directory, gpu)

    # Get latest checkpoint
    checkpoint = torchutil.checkpoint.latest_path(directory)

    # Evaluate
    penn.evaluate.datasets(penn.EVALUATION_DATASETS, checkpoint, gpu)


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(description='Train a model')
    parser.add_argument(
        '--config',
        type=Path,
        required=True,
        help='The configuration file')
    parser.add_argument(
        '--datasets',
        nargs='+',
        default=penn.DATASETS,
        help='The datasets to train on')
    parser.add_argument(
        '--gpu',
        type=int,
        help='The GPU index')
    return parser.parse_args()


main(**vars(parse_args()))



import functools

import torch
import torchutil

import penn


###############################################################################
# Training
###############################################################################


@torchutil.notify('train')
def train(datasets, directory, gpu=None):
    """Train a model"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')
    model = penn.model.Model().to(device)

    #######################
    # Create data loaders #
    #######################

    torch.manual_seed(penn.RANDOM_SEED)
    train_loader = penn.data.loader(datasets, 'train')
    valid_loader = penn.data.loader(datasets, 'valid')

    ####################
    # Create optimizer #
    ####################

    optimizer = torch.optim.Adam(model.parameters(), lr=penn.LEARNING_RATE)

    ##############################
    # Maybe load from checkpoint #
    ##############################

    path = torchutil.checkpoint.latest_path(directory)

    if path is not None:

        # Load model
        model, optimizer, state = torchutil.checkpoint.load(
            path,
            model,
            optimizer)
        step, epoch = state['step'], state['epoch']

    else:

        # Train from scratch
        step, epoch = 0, 0

    ##############################
    # Maybe setup early stopping #
    ##############################

    if penn.EARLY_STOPPING:
        counter = penn.EARLY_STOPPING_STEPS
        best_accuracy = 0.
        stop = False

    #########
    # Train #
    #########

    # Automatic mixed precision (amp) gradient scaler
    scaler = torch.cuda.amp.GradScaler()

    # Setup progress bar
    progress = torchutil.iterator(
        range(step, penn.STEPS),
        f'Training {penn.CONFIG}',
        step,
        penn.STEPS)
    while step < penn.STEPS and (not penn.EARLY_STOPPING or not stop):

        # Seed sampler
        train_loader.sampler.set_epoch(epoch)

        for batch in train_loader:

            # Unpack batch
            audio, bins, *_ = batch

            with torch.autocast(device.type):

                # Forward pass
                logits = model(audio.to(device))

                # Compute losses
                losses = loss(logits, bins.to(device))

            ##################
            # Optimize model #
            ##################

            optimizer.zero_grad()

            # Backward pass
            scaler.scale(losses).backward()

            # Update weights
            scaler.step(optimizer)

            # Update gradient scaler
            scaler.update()

            ##############
            # Evaluation #
            ##############

            # Save checkpoint
            if step and step % penn.CHECKPOINT_INTERVAL == 0:
                torchutil.checkpoint.save(
                    directory / f'{step:08d}.pt',
                    model,
                    optimizer,
                    step=step,
                    epoch=epoch)

            # Evaluate
            if step % penn.LOG_INTERVAL == 0:
                evaluate_fn = functools.partial(
                    evaluate,
                    directory,
                    step,
                    model,
                    gpu)
                evaluate_fn('train', train_loader)
                valid_accuracy = evaluate_fn('valid', valid_loader)

                # Maybe stop training
                if penn.EARLY_STOPPING:
                    counter -= 1

                    # Update best validation loss
                    if valid_accuracy > best_accuracy:
                        best_accuracy = valid_accuracy
                        counter = penn.EARLY_STOPPING_STEPS

                    # Stop training
                    elif counter == 0:
                        stop = True

            # Update training step count
            if step >= penn.STEPS or (penn.EARLY_STOPPING and stop):
                break
            step += 1

            # Update progress bar
            progress.update()

        # Update epoch
        epoch += 1

    # Close progress bar
    progress.close()

    # Save final model
    torchutil.checkpoint.save(
        directory / f'{step:08d}.pt',
        model,
        optimizer,
        step=step,
        epoch=epoch)


###############################################################################
# Evaluation
###############################################################################


def evaluate(directory, step, model, gpu, condition, loader):
    """Perform model evaluation"""
    device = torch.device('cpu' if gpu is None else f'cuda:{gpu}')

    # Setup evaluation metrics
    metrics = penn.evaluate.Metrics()

    # Prepare model for inference
    with penn.inference_context(model):

        # Unpack batch
        for i, (audio, bins, pitch, voiced, *_) in enumerate(loader):

            # Forward pass
            logits = model(audio.to(device))

            # Update metrics
            metrics.update(
                logits.to(device),
                bins.T.to(device),
                pitch.T.to(device),
                voiced.T.to(device))

            # Stop when we exceed some number of batches
            if i + 1 == penn.LOG_STEPS:
                break

    # Format results
    scalars = {
        f'{key}/{condition}': value for key, value in metrics().items()}

    # Write to tensorboard
    torchutil.tensorboard.update(directory, step, scalars=scalars)

    return scalars[f'accuracy/{condition}']


###############################################################################
# Loss function
###############################################################################


def loss(logits, bins):
    """Compute loss function"""
    # Reshape inputs
    logits = logits.permute(0, 2, 1).reshape(-1, penn.PITCH_BINS)
    bins = bins.flatten()

    # Maybe blur target
    if penn.GAUSSIAN_BLUR:

        # Cache cents values to evaluate distributions at
        if not hasattr(loss, 'cents'):
            loss.cents = penn.convert.bins_to_cents(
                torch.arange(penn.PITCH_BINS))[:, None]

        # Ensure values are on correct device (no-op if devices are the same)
        loss.cents = loss.cents.to(bins.device)

        # Create normal distributions
        distributions = torch.distributions.Normal(
            penn.convert.bins_to_cents(bins),
            25)

        # Sample normal distributions
        bins = torch.exp(distributions.log_prob(loss.cents)).permute(1, 0)

        # Normalize
        bins = bins / (bins.max(dim=1, keepdims=True).values + 1e-8)

    else:

        # One-hot encoding
        bins = torch.nn.functional.one_hot(bins, penn.PITCH_BINS).float()

    if penn.LOSS == 'binary_cross_entropy':

        # Compute binary cross-entropy loss
        return torch.nn.functional.binary_cross_entropy_with_logits(
            logits,
            bins)

    elif penn.LOSS == 'categorical_cross_entropy':

        # Compute categorical cross-entropy loss
        return torch.nn.functional.cross_entropy(logits, bins)

    else:

        raise ValueError(f'Loss {penn.LOSS} is not implemented')



import shutil

import torchutil

import penn


###############################################################################
# Download datasets
###############################################################################


@torchutil.notify('download')
def datasets(datasets):
    """Download datasets"""
    if 'mdb' in datasets:
        mdb()

    if 'ptdb' in datasets:
        ptdb()


###############################################################################
# Individual datasets
###############################################################################


def mdb():
    """Download mdb dataset"""
    torchutil.download.targz(
        'https://zenodo.org/record/1481172/files/MDB-stem-synth.tar.gz',
        penn.DATA_DIR)

    # Delete previous directory
    shutil.rmtree(penn.DATA_DIR / 'mdb', ignore_errors=True)

    # Rename directory
    shutil.move(penn.DATA_DIR / 'MDB-stem-synth', penn.DATA_DIR / 'mdb')


def ptdb():
    """Download ptdb dataset"""
    directory = penn.DATA_DIR / 'ptdb'
    directory.mkdir(exist_ok=True, parents=True)
    torchutil.download.zip(
        'https://www2.spsc.tugraz.at/databases/PTDB-TUG/SPEECH_DATA_ZIPPED.zip',
        directory)



import json
import warnings

import torchaudio

import penn


def audio(file):
    """Load audio from disk"""
    audio, sample_rate = torchaudio.load(file)

    # If audio is stereo, convert to mono
    if audio.size(0) == 2:
        warnings.warn(f'Converting stereo audio to mono: {file}')
        audio = audio.mean(dim=0, keepdim=True)

    # Maybe resample
    return penn.resample(audio, sample_rate)


def partition(dataset):
    """Load partitions for dataset"""
    with open(penn.PARTITION_DIR / f'{dataset}.json') as file:
        return json.load(file)
